import math
from typing import Any

import torch
from torch.nn import Parameter
from torch.nn.modules import Conv2d, Module


class GaborConv2d(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=False,
        padding_mode="zeros",
    ):
        super().__init__()

        self.is_calculated = False

        self.conv_layer = Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.kernel_size = self.conv_layer.kernel_size

        # small addition to avoid division by zero
        self.delta = 1e-3

        # freq, theta, sigma are set up according to S. Meshgini,
        # A. Aghagolzadeh and H. Seyedarabi, "Face recognition using
        # Gabor filter bank, kernel principal component analysis
        # and support vector machine"
        self.freq = Parameter(
            (math.pi / 2)
            * math.sqrt(2)
            ** (-torch.randint(0, 5, (out_channels, in_channels))).type(torch.Tensor),
            requires_grad=True,
        )
        self.theta = Parameter(
            (math.pi / 8)
            * torch.randint(0, 8, (out_channels, in_channels)).type(torch.Tensor),
            requires_grad=True,
        )
        self.sigma = Parameter(math.pi / self.freq, requires_grad=True)
        self.psi = Parameter(
            math.pi * torch.rand(out_channels, in_channels), requires_grad=True
        )

        self.x0 = Parameter(
            torch.ceil(torch.Tensor([self.kernel_size[0] / 2]))[0], requires_grad=False
        )
        self.y0 = Parameter(
            torch.ceil(torch.Tensor([self.kernel_size[1] / 2]))[0], requires_grad=False
        )

        self.y, self.x = torch.meshgrid(
            [
                torch.linspace(-self.x0 + 1, self.x0 + 0, self.kernel_size[0]),
                torch.linspace(-self.y0 + 1, self.y0 + 0, self.kernel_size[1]),
            ]
        )
        self.y = Parameter(self.y)
        self.x = Parameter(self.x)

        self.weight = Parameter(
            torch.empty(self.conv_layer.weight.shape, requires_grad=True),
            requires_grad=True,
        )

        self.register_parameter("freq", self.freq)
        self.register_parameter("theta", self.theta)
        self.register_parameter("sigma", self.sigma)
        self.register_parameter("psi", self.psi)
        self.register_parameter("x_shape", self.x0)
        self.register_parameter("y_shape", self.y0)
        self.register_parameter("y_grid", self.y)
        self.register_parameter("x_grid", self.x)
        self.register_parameter("weight", self.weight)

    def forward(self, input_tensor):
        if self.training:
            self.calculate_weights()
            self.is_calculated = False
        if not self.training:
            if not self.is_calculated:
                self.calculate_weights()
                self.is_calculated = True
        return self.conv_layer(input_tensor)

    def calculate_weights(self):
        for i in range(self.conv_layer.out_channels):
            for j in range(self.conv_layer.in_channels):
                sigma = self.sigma[i, j].expand_as(self.y)
                freq = self.freq[i, j].expand_as(self.y)
                theta = self.theta[i, j].expand_as(self.y)
                psi = self.psi[i, j].expand_as(self.y)

                rotx = self.x * torch.cos(theta) + self.y * torch.sin(theta)
                roty = -self.x * torch.sin(theta) + self.y * torch.cos(theta)

                g = torch.exp(
                    -0.5 * ((rotx ** 2 + roty ** 2) / (sigma + self.delta) ** 2)
                )
                g = g * torch.cos(freq * rotx + psi)
                g = g / (2 * math.pi * sigma ** 2)
                self.conv_layer.weight.data[i, j] = g

    def _forward_unimplemented(self, *inputs: Any):
        """
        code checkers makes implement this method,
        looks like error in PyTorch
        """
        raise NotImplementedError



import numpy as np
import torch

from GaborNet import GaborConv2d


def test_kenrel_shape():
    k = np.random.randint(11, 15)
    layer = GaborConv2d(1, 1, (k, k))
    shape = np.array(layer.weight.shape)
    result = k
    assert result == shape[-1]


def test_output_size():
    s = np.random.randint(3, 9)
    layer = GaborConv2d(1, s, (11, 11))
    shape = np.array(layer.weight.shape)
    result = s
    assert result == shape[0]


def test_input_size():
    s = np.random.randint(3, 9)
    layer = GaborConv2d(s, 1, (11, 11))
    shape = np.array(layer.weight.shape)
    result = s
    assert result == shape[1]


def test_layer_work():
    img = np.random.randn(1, 1, 100, 100)
    img = torch.Tensor(img)
    k = np.random.randint(11, 15)
    layer = GaborConv2d(1, 1, (k, k))
    output = layer(img)
    shape = np.array(output.shape)
    result = 100 - k + 1
    assert result == shape[-1] == shape[-2]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['GaborNet', 'GaborNet.tests']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.1,<2.0.0', 'torch>=1.6.0,<2.0.0', 'torchvision>=0.7.0,<0.8.0']

setup_kwargs = {
    'name': 'gabornet',
    'version': '0.2.0',
    'description': 'Meet Gabor Layer',
    'long_description': '# GaborNet\n\n[![PyPI-Status][pypi-image]][pypi-url]\n[![Build Status][travis-badge]][travis-url]\n[![LICENSE][license-image]][license-url]\n[![DeepSource](https://static.deepsource.io/deepsource-badge-light-mini.svg)](https://deepsource.io/gh/iKintosh/GaborNet/?ref=repository-badge)\n\n## Installation\n\nGaborNet can be installed via pip from Python 3.7 and above:\n\n```bash\npip install GaborNet\n```\n\n## Getting started\n\n```python\nimport torch\nimport torch.nn as nn\nfrom torch.nn import functional as F\nfrom GaborNet import GaborConv2d\n\ndevice = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")\n\n\nclass GaborNN(nn.Module):\n    def __init__(self):\n        super(GaborNN, self).__init__()\n        self.g0 = GaborConv2d(in_channels=1, out_channels=96, kernel_size=(11, 11))\n        self.c1 = nn.Conv2d(96, 384, (3,3))\n        self.fc1 = nn.Linear(384*3*3, 64)\n        self.fc2 = nn.Linear(64, 2)\n\n    def forward(self, x):\n        x = F.leaky_relu(self.g0(x))\n        x = nn.MaxPool2d()(x)\n        x = F.leaky_relu(self.c1(x))\n        x = nn.MaxPool2d()(x)\n        x = x.view(-1, 384*3*3)\n        x = F.leaky_relu(self.fc1(x))\n        x = self.fc2(x)\n        return x\n\nnet = GaborNN().to(device)\n\n```\n\nOriginal research paper (preprint): https://arxiv.org/abs/1904.13204\n\nThis research on deep convolutional neural networks proposes a modified\narchitecture that focuses on improving convergence and reducing training\ncomplexity. The filters in the first layer of network are constrained to fit the\nGabor function. The parameters of Gabor functions are learnable and updated by\nstandard backpropagation techniques. The proposed architecture was tested on\nseveral datasets and outperformed the common convolutional networks\n\n## Citation\n\nPlease use this bibtex if you want to cite this repository in your publications:\n\n    @misc{gabornet,\n        author = {Alekseev, Andrey},\n        title = {GaborNet: Gabor filters with learnable parameters in deep convolutional\n                   neural networks},\n        year = {2019},\n        publisher = {GitHub},\n        journal = {GitHub repository},\n        howpublished = {\\url{https://github.com/iKintosh/GaborNet}},\n    }\n\n[travis-url]: https://travis-ci.com/iKintosh/GaborNet\n[travis-badge]: https://travis-ci.com/iKintosh/GaborNet.svg?branch=master\n[pypi-image]: https://img.shields.io/pypi/v/gabornet.svg\n[pypi-url]: https://pypi.org/project/gabornet\n[license-image]: https://img.shields.io/badge/License-MIT-yellow.svg\n[license-url]: https://pypi.org/project/gabornet\n',
    'author': 'an.alekseev',
    'author_email': 'alekseev.as@phystech.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iKintosh/GaborNet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
